#include "sm3sum.h"

#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include <ctype.h>

#define BUF_LEN 4096

static size_t get_file_size(char * filename)
{
	FILE * fin = fopen(filename, "rb");
	if (!fin) return 0;

	fseek( fin, 0, SEEK_END );
	size_t f_length = ftell(fin);
	fclose(fin);
	return f_length;
}

static u4 swap_endian(u4 data)
{
    u4 ret;
    ret = (data >> 24) |
          ((data<<8) & 0x00FF0000) |
          ((data>>8) & 0x0000FF00) |
          (data << 24);
    return ret;
}

/*
 * convert sm3 string to int array
 * sm3_str: char array that contains sm3 string
 * ret: u4 array of size 32 == (sizeof(u4) * 8)
 */
void sm3str_to_int(char * sm3_str, u4 * ret)
{
    for (size_t i = 0; i < 8; i++)
    {
        u4 result = 0;
        for (size_t j = 0; j < 8; j++)
        {
            u4 tmp = tolower(sm3_str[i * 8 + j]);
            result = result << 4 | (isdigit(tmp) ? tmp - '0' : tmp - 'a' + 10);
        }
        ret[i] = swap_endian(result);
    }
}


/*
 * fp: the FILE pointer of data to be calculated
 * fileSize: the file size in bits
 * buf: u1 array, block size: 36MB + 1 block (for padding) == u1 buf[(BLOCK_BATCH_CNT * 2) * BLOCK_SIZE];
 */
static void read_and_calc(FILE * fp, size_t fileSize, u4 * hash)
{
    size_t calculated = 0;
    size_t readLen;
	
	u1 sm3buf[BUF_LEN];
	memset( sm3buf, 0, sizeof(sm3buf) );

	SM3_CTX ctx;
    sm3_init(&ctx);
    
	while (calculated + BUF_LEN <= fileSize)
    {
        readLen = fread(sm3buf, 1, BUF_LEN, fp);
        sm3_update(&ctx, sm3buf, readLen);
        calculated += readLen;
    }

    // may not occupy full space, needs clean up
    memset( sm3buf, 0, sizeof(sm3buf) );
    size_t finalSize = fileSize - calculated;
    readLen = fread(sm3buf, 1, finalSize, fp);
    sm3_update(&ctx, sm3buf, readLen);
    
	sm3_final(&ctx, (u1 *)hash);
}

static void stdin_read_and_calc(u4 * hash)
{
	u1 sm3buf[BUF_LEN];
	memset( sm3buf, 0, sizeof(sm3buf) );

	SM3_CTX ctx;
    sm3_init(&ctx);

    size_t offset = 0;
    int c = getchar();
    while (c != EOF)
    {
        sm3buf[offset++] = c;
        c = getchar();
        if (offset == BUF_LEN && c != EOF)
        {
            offset = 0;
            sm3_update(&ctx, sm3buf, BUF_LEN);
        }
    }
    
    // memset(sm3buf + offset, 0, BUF_LEN - offset);
    sm3_update(&ctx, sm3buf, offset);
    sm3_final(&ctx, (u1 *)hash);
}


void calc_file_hash(sm3_file_ctx * fileCTX)
{
	if (strncmp(fileCTX->filename, "-", 2) == 0)
	{
		stdin_read_and_calc(fileCTX->calculated_sm3);
		return;
	}

	size_t fileSize;
	fileSize = get_file_size(fileCTX->filename);

	FILE * targetFile;
	targetFile = fopen(fileCTX->filename, "rb");
	if (targetFile == NULL)
	{
		printf("Cannot access file %s, either non-existing or not readable\n", fileCTX->filename);
		return;
	}
	
	// todo: output results
	read_and_calc(targetFile, fileSize, fileCTX->calculated_sm3);
	fclose(targetFile);
}


/*
 * calculate hash for every file specified
 * only read from the list with head hash_pair_head
 */
int check_file_hash(sm3_file_ctx * fileCTX)
{
	calc_file_hash(fileCTX);
	
	if (memcmp(fileCTX->expected_sm3, fileCTX->calculated_sm3, 256/8) == 0) // not constant time comparison!
	{
		printf("%s: OK\n", fileCTX->filename);
		return 0;
	}
	
	printf("%s: FAILED, correct checksum should be:\n", fileCTX->filename);
	print_file_sm3(fileCTX);
	
	return -1;
}

void print_file_sm3(sm3_file_ctx * ctx)
{
    for (size_t i = 0; i < 8; i++)
    {
        printf("%08x", swap_endian(ctx->calculated_sm3[i]));
    }

	printf(" %s\n", ctx->filename);
}
